@@ -25,13 +25,13 @@
 #include "shrpx_client_handler.h"
 
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#  include <unistd.h>
 #endif // HAVE_UNISTD_H
 #ifdef HAVE_SYS_SOCKET_H
-#include <sys/socket.h>
+#  include <sys/socket.h>
 #endif // HAVE_SYS_SOCKET_H
 #ifdef HAVE_NETDB_H
-#include <netdb.h>
+#  include <netdb.h>
 #endif // HAVE_NETDB_H
 
 #include <cerrno>
@@ -111,6 +111,7 @@ void writecb(struct ev_loop *loop, ev_io *w, int revents) {
 int ClientHandler::noop() { return 0; }
 
 int ClientHandler::read_clear() {
+  auto should_break = false;
   rb_.ensure_chunk();
   for (;;) {
     if (rb_.rleft() && on_read() != 0) {
@@ -123,7 +124,7 @@ int ClientHandler::read_clear() {
       return 0;
     }
 
-    if (!ev_is_active(&conn_.rev)) {
+    if (!ev_is_active(&conn_.rev) || should_break) {
       return 0;
     }
 
@@ -141,6 +142,7 @@ int ClientHandler::read_clear() {
     }
 
     rb_.write(nread);
+    should_break = true;
   }
 }
 
@@ -205,6 +207,8 @@ int ClientHandler::tls_handshake() {
 }
 
 int ClientHandler::read_tls() {
+  auto should_break = false;
+
   ERR_clear_error();
 
   rb_.ensure_chunk();
@@ -221,7 +225,7 @@ int ClientHandler::read_tls() {
       return 0;
     }
 
-    if (!ev_is_active(&conn_.rev)) {
+    if (!ev_is_active(&conn_.rev) || should_break) {
       return 0;
     }
 
@@ -239,6 +243,7 @@ int ClientHandler::read_tls() {
     }
 
     rb_.write(nread);
+    should_break = true;
   }
 }
 
@@ -396,7 +401,7 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
             get_config()->conn.upstream.ratelimit.write,
             get_config()->conn.upstream.ratelimit.read, writecb, readcb,
             timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
-            get_config()->tls.dyn_rec.idle_timeout, PROTO_NONE),
+            get_config()->tls.dyn_rec.idle_timeout, Proto::NONE),
       ipaddr_(make_string_ref(balloc_, ipaddr)),
       port_(make_string_ref(balloc_, port)),
       faddr_(faddr),
@@ -430,7 +435,7 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
   auto &fwdconf = config->http.forwarded;
 
   if (fwdconf.params & FORWARDED_FOR) {
-    if (fwdconf.for_node_type == FORWARDED_NODE_OBFUSCATED) {
+    if (fwdconf.for_node_type == ForwardedNode::OBFUSCATED) {
       // 1 for '_'
       auto len = SHRPX_OBFUSCATED_NODE_LENGTH + 1;
       // 1 for terminating NUL.
@@ -478,7 +483,7 @@ void ClientHandler::setup_upstream_io_callback() {
     // For non-TLS version, first create HttpsUpstream. It may be
     // upgraded to HTTP/2 through HTTP Upgrade or direct HTTP/2
     // connection.
-    upstream_ = make_unique<HttpsUpstream>(this);
+    upstream_ = std::make_unique<HttpsUpstream>(this);
     alpn_ = StringRef::from_lit("http/1.1");
     read_ = &ClientHandler::read_clear;
     write_ = &ClientHandler::write_clear;
@@ -549,7 +554,9 @@ int ClientHandler::validate_next_proto() {
   // First set callback for catch all cases
   on_read_ = &ClientHandler::upstream_read;
 
+#ifndef OPENSSL_NO_NEXTPROTONEG
   SSL_get0_next_proto_negotiated(conn_.tls.ssl, &next_proto, &next_proto_len);
+#endif // !OPENSSL_NO_NEXTPROTONEG
 #if OPENSSL_VERSION_NUMBER >= 0x10002000L
   if (next_proto == nullptr) {
     SSL_get0_alpn_selected(conn_.tls.ssl, &next_proto, &next_proto_len);
@@ -582,7 +589,7 @@ int ClientHandler::validate_next_proto() {
   if (util::check_h2_is_selected(proto)) {
     on_read_ = &ClientHandler::upstream_http2_connhd_read;
 
-    auto http2_upstream = make_unique<Http2Upstream>(this);
+    auto http2_upstream = std::make_unique<Http2Upstream>(this);
 
     upstream_ = std::move(http2_upstream);
     alpn_ = make_string_ref(balloc_, proto);
@@ -598,7 +605,7 @@ int ClientHandler::validate_next_proto() {
   }
 
   if (proto == StringRef::from_lit("http/1.1")) {
-    upstream_ = make_unique<HttpsUpstream>(this);
+    upstream_ = std::make_unique<HttpsUpstream>(this);
     alpn_ = StringRef::from_lit("http/1.1");
 
     // At this point, input buffer is already filled with some bytes.
@@ -656,30 +663,11 @@ void ClientHandler::pool_downstream_connection(
                      << " in group " << group;
   }
 
-  auto &shared_addr = group->shared_addr;
-
-  if (shared_addr->affinity.type == AFFINITY_NONE) {
-    auto &dconn_pool = group->shared_addr->dconn_pool;
-    dconn_pool.add_downstream_connection(std::move(dconn));
-
-    return;
-  }
-
   auto addr = dconn->get_addr();
   auto &dconn_pool = addr->dconn_pool;
   dconn_pool->add_downstream_connection(std::move(dconn));
 }
 
-void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
-  if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Removing downstream connection DCONN:" << dconn
-                     << " from pool";
-  }
-  auto &dconn_pool =
-      dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
-  dconn_pool.remove_downstream_connection(dconn);
-}
-
 namespace {
 // Computes 32bits hash for session affinity for IP address |ip|.
 uint32_t compute_affinity_from_ip(const StringRef &ip) {
@@ -699,7 +687,7 @@ uint32_t compute_affinity_from_ip(const StringRef &ip) {
 }
 } // namespace
 
-Http2Session *ClientHandler::select_http2_session_with_affinity(
+Http2Session *ClientHandler::get_http2_session(
     const std::shared_ptr<DownstreamAddrGroup> &group, DownstreamAddr *addr) {
   auto &shared_addr = group->shared_addr;
 
@@ -752,171 +740,6 @@ Http2Session *ClientHandler::select_http2_session_with_affinity(
   return session;
 }
 
-namespace {
-// Returns true if load of |lhs| is lighter than that of |rhs|.
-// Currently, we assume that lesser streams means lesser load.
-bool load_lighter(const DownstreamAddr *lhs, const DownstreamAddr *rhs) {
-  return lhs->num_dconn < rhs->num_dconn;
-}
-} // namespace
-
-Http2Session *ClientHandler::select_http2_session(
-    const std::shared_ptr<DownstreamAddrGroup> &group) {
-  auto &shared_addr = group->shared_addr;
-
-  // First count the working backend addresses.
-  size_t min = 0;
-  for (const auto &addr : shared_addr->addrs) {
-    if (addr.proto != PROTO_HTTP2 || addr.connect_blocker->blocked()) {
-      continue;
-    }
-
-    ++min;
-  }
-
-  if (min == 0) {
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "No working backend address found";
-    }
-
-    return nullptr;
-  }
-
-  auto &http2_avail_freelist = shared_addr->http2_avail_freelist;
-
-  if (http2_avail_freelist.size() >= min) {
-    for (auto session = http2_avail_freelist.head; session;) {
-      auto next = session->dlnext;
-
-      session->remove_from_freelist();
-
-      // session may be in graceful shutdown period now.
-      if (session->max_concurrency_reached(0)) {
-        if (LOG_ENABLED(INFO)) {
-          CLOG(INFO, this)
-              << "Maximum streams have been reached for Http2Session("
-              << session << ").  Skip it";
-        }
-
-        session = next;
-
-        continue;
-      }
-
-      if (LOG_ENABLED(INFO)) {
-        CLOG(INFO, this) << "Use Http2Session " << session
-                         << " from http2_avail_freelist";
-      }
-
-      if (session->max_concurrency_reached(1)) {
-        if (LOG_ENABLED(INFO)) {
-          CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
-                           << session << ").";
-        }
-      } else {
-        session->add_to_avail_freelist();
-      }
-      return session;
-    }
-  }
-
-  DownstreamAddr *selected_addr = nullptr;
-
-  for (auto &addr : shared_addr->addrs) {
-    if (addr.in_avail || addr.proto != PROTO_HTTP2 ||
-        (addr.http2_extra_freelist.size() == 0 &&
-         addr.connect_blocker->blocked())) {
-      continue;
-    }
-
-    for (auto session = addr.http2_extra_freelist.head; session;) {
-      auto next = session->dlnext;
-
-      // session may be in graceful shutdown period now.
-      if (session->max_concurrency_reached(0)) {
-        if (LOG_ENABLED(INFO)) {
-          CLOG(INFO, this)
-              << "Maximum streams have been reached for Http2Session("
-              << session << ").  Skip it";
-        }
-
-        session->remove_from_freelist();
-
-        session = next;
-
-        continue;
-      }
-
-      break;
-    }
-
-    if (selected_addr == nullptr || load_lighter(&addr, selected_addr)) {
-      selected_addr = &addr;
-    }
-  }
-
-  assert(selected_addr);
-
-  if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Selected DownstreamAddr=" << selected_addr
-                     << ", index="
-                     << (selected_addr - shared_addr->addrs.data());
-  }
-
-  if (selected_addr->http2_extra_freelist.size()) {
-    auto session = selected_addr->http2_extra_freelist.head;
-    session->remove_from_freelist();
-
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Use Http2Session " << session
-                       << " from http2_extra_freelist";
-    }
-
-    if (session->max_concurrency_reached(1)) {
-      if (LOG_ENABLED(INFO)) {
-        CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
-                         << session << ").";
-      }
-    } else {
-      session->add_to_avail_freelist();
-    }
-    return session;
-  }
-
-  auto session = new Http2Session(conn_.loop, worker_->get_cl_ssl_ctx(),
-                                  worker_, group, selected_addr);
-
-  if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Create new Http2Session " << session;
-  }
-
-  session->add_to_avail_freelist();
-
-  return session;
-}
-
-namespace {
-// The chosen value is small enough for uint32_t, and large enough for
-// the number of backend.
-constexpr uint32_t WEIGHT_MAX = 65536;
-} // namespace
-
-namespace {
-bool pri_less(const WeightedPri &lhs, const WeightedPri &rhs) {
-  if (lhs.cycle < rhs.cycle) {
-    return rhs.cycle - lhs.cycle <= WEIGHT_MAX;
-  }
-
-  return lhs.cycle - rhs.cycle > WEIGHT_MAX;
-}
-} // namespace
-
-namespace {
-uint32_t next_cycle(const WeightedPri &pri) {
-  return pri.cycle + WEIGHT_MAX / std::min(WEIGHT_MAX, pri.weight);
-}
-} // namespace
-
 uint32_t ClientHandler::get_affinity_cookie(Downstream *downstream,
                                             const StringRef &cookie_name) {
   auto h = downstream->find_affinity_cookie(cookie_name);
@@ -935,82 +758,60 @@ uint32_t ClientHandler::get_affinity_cookie(Downstream *downstream,
   return h;
 }
 
-std::unique_ptr<DownstreamConnection>
-ClientHandler::get_downstream_connection(int &err, Downstream *downstream) {
-  size_t group_idx;
-  auto &downstreamconf = *worker_->get_downstream_config();
-  auto &routerconf = downstreamconf.router;
+namespace {
+void reschedule_addr(
+    std::priority_queue<DownstreamAddrEntry, std::vector<DownstreamAddrEntry>,
+                        DownstreamAddrEntryGreater> &pq,
+    DownstreamAddr *addr) {
+  auto penalty = MAX_DOWNSTREAM_ADDR_WEIGHT + addr->pending_penalty;
+  addr->cycle += penalty / addr->weight;
+  addr->pending_penalty = penalty % addr->weight;
 
-  auto catch_all = downstreamconf.addr_group_catch_all;
-  auto &groups = worker_->get_downstream_addr_groups();
+  pq.push(DownstreamAddrEntry{addr, addr->seq, addr->cycle});
+  addr->queued = true;
+}
+} // namespace
 
-  const auto &req = downstream->request();
+namespace {
+void reschedule_wg(
+    std::priority_queue<WeightGroupEntry, std::vector<WeightGroupEntry>,
+                        WeightGroupEntryGreater> &pq,
+    WeightGroup *wg) {
+  auto penalty = MAX_DOWNSTREAM_ADDR_WEIGHT + wg->pending_penalty;
+  wg->cycle += penalty / wg->weight;
+  wg->pending_penalty = penalty % wg->weight;
 
+  pq.push(WeightGroupEntry{wg, wg->seq, wg->cycle});
+  wg->queued = true;
+}
+} // namespace
+
+DownstreamAddr *ClientHandler::get_downstream_addr(int &err,
+                                                   DownstreamAddrGroup *group,
+                                                   Downstream *downstream) {
   err = 0;
 
   switch (faddr_->alt_mode) {
-  case ALTMODE_API:
-    return make_unique<APIDownstreamConnection>(worker_);
-  case ALTMODE_HEALTHMON:
-    return make_unique<HealthMonitorDownstreamConnection>();
+  case UpstreamAltMode::API:
+  case UpstreamAltMode::HEALTHMON:
+    assert(0);
+  default:
+    break;
   }
 
-  auto &balloc = downstream->get_block_allocator();
-
-  // Fast path.  If we have one group, it must be catch-all group.
-  if (groups.size() == 1) {
-    group_idx = 0;
-  } else {
-    StringRef authority;
-    if (faddr_->sni_fwd) {
-      authority = sni_;
-    } else if (!req.authority.empty()) {
-      authority = req.authority;
-    } else {
-      auto h = req.fs.header(http2::HD_HOST);
-      if (h) {
-        authority = h->value;
-      }
-    }
-
-    StringRef path;
-    // CONNECT method does not have path.  But we requires path in
-    // host-path mapping.  As workaround, we assume that path is "/".
-    if (req.method != HTTP_CONNECT) {
-      path = req.path;
-    }
-
-    group_idx = match_downstream_addr_group(routerconf, authority, path, groups,
-                                            catch_all, balloc);
-  }
-
-  if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Downstream address group_idx: " << group_idx;
-  }
-
-  if (groups[group_idx]->shared_addr->redirect_if_not_tls && !conn_.tls.ssl) {
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Downstream address group " << group_idx
-                       << " requires frontend TLS connection.";
-    }
-    err = SHRPX_ERR_TLS_REQUIRED;
-    return nullptr;
-  }
-
-  auto &group = groups[group_idx];
   auto &shared_addr = group->shared_addr;
 
-  if (shared_addr->affinity.type != AFFINITY_NONE) {
+  if (shared_addr->affinity.type != SessionAffinity::NONE) {
     uint32_t hash;
     switch (shared_addr->affinity.type) {
-    case AFFINITY_IP:
+    case SessionAffinity::IP:
       if (!affinity_hash_computed_) {
         affinity_hash_ = compute_affinity_from_ip(ipaddr_);
         affinity_hash_computed_ = true;
       }
       hash = affinity_hash_;
       break;
-    case AFFINITY_COOKIE:
+    case SessionAffinity::COOKIE:
       hash = get_affinity_cookie(downstream, shared_addr->affinity.cookie.name);
       break;
     default:
@@ -1051,101 +852,153 @@ ClientHandler::get_downstream_connection(int &err, Downstream *downstream) {
       aff_idx = i;
     }
 
-    if (addr->proto == PROTO_HTTP2) {
-      auto http2session = select_http2_session_with_affinity(group, addr);
-
-      auto dconn = make_unique<Http2DownstreamConnection>(http2session);
-
-      dconn->set_client_handler(this);
-
-      return std::move(dconn);
-    }
-
-    auto &dconn_pool = addr->dconn_pool;
-    auto dconn = dconn_pool->pop_downstream_connection();
-
-    if (!dconn) {
-      dconn = make_unique<HttpDownstreamConnection>(group, aff_idx, conn_.loop,
-                                                    worker_);
-    }
-
-    dconn->set_client_handler(this);
-
-    return dconn;
+    return addr;
   }
 
-  auto http1_weight = shared_addr->http1_pri.weight;
-  auto http2_weight = shared_addr->http2_pri.weight;
+  auto &wgpq = shared_addr->pq;
 
-  auto proto = PROTO_NONE;
-
-  if (http1_weight > 0 && http2_weight > 0) {
-    // We only advance cycle if both weight has nonzero to keep its
-    // distance under WEIGHT_MAX.
-    if (pri_less(shared_addr->http1_pri, shared_addr->http2_pri)) {
-      proto = PROTO_HTTP1;
-      shared_addr->http1_pri.cycle = next_cycle(shared_addr->http1_pri);
-    } else {
-      proto = PROTO_HTTP2;
-      shared_addr->http2_pri.cycle = next_cycle(shared_addr->http2_pri);
-    }
-  } else if (http1_weight > 0) {
-    proto = PROTO_HTTP1;
-  } else if (http2_weight > 0) {
-    proto = PROTO_HTTP2;
-  }
-
-  if (proto == PROTO_NONE) {
-    if (LOG_ENABLED(INFO)) {
+  for (;;) {
+    if (wgpq.empty()) {
       CLOG(INFO, this) << "No working downstream address found";
-    }
-
-    err = -1;
-    return nullptr;
-  }
-
-  if (proto == PROTO_HTTP2) {
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Downstream connection pool is empty."
-                       << " Create new one";
-    }
-
-    auto http2session = select_http2_session(group);
-
-    if (http2session == nullptr) {
       err = -1;
       return nullptr;
     }
 
-    auto dconn = make_unique<Http2DownstreamConnection>(http2session);
+    auto wg = wgpq.top().wg;
+    wgpq.pop();
+    wg->queued = false;
 
-    dconn->set_client_handler(this);
+    for (;;) {
+      if (wg->pq.empty()) {
+        break;
+      }
 
-    return std::move(dconn);
+      auto addr = wg->pq.top().addr;
+      wg->pq.pop();
+      addr->queued = false;
+
+      if (addr->connect_blocker->blocked()) {
+        continue;
+      }
+
+      reschedule_addr(wg->pq, addr);
+      reschedule_wg(wgpq, wg);
+
+      return addr;
+    }
+  }
+}
+
+std::unique_ptr<DownstreamConnection>
+ClientHandler::get_downstream_connection(int &err, Downstream *downstream) {
+  size_t group_idx;
+  auto &downstreamconf = *worker_->get_downstream_config();
+  auto &routerconf = downstreamconf.router;
+
+  auto catch_all = downstreamconf.addr_group_catch_all;
+  auto &groups = worker_->get_downstream_addr_groups();
+
+  auto &req = downstream->request();
+
+  err = 0;
+
+  switch (faddr_->alt_mode) {
+  case UpstreamAltMode::API:
+    return std::make_unique<APIDownstreamConnection>(worker_);
+  case UpstreamAltMode::HEALTHMON:
+    return std::make_unique<HealthMonitorDownstreamConnection>();
+  default:
+    break;
   }
 
-  auto &dconn_pool = shared_addr->dconn_pool;
+  auto &balloc = downstream->get_block_allocator();
 
-  // pool connection must be HTTP/1.1 connection
-  auto dconn = dconn_pool.pop_downstream_connection();
+  StringRef authority, path;
 
-  if (dconn) {
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Reuse downstream connection DCONN:" << dconn.get()
-                       << " from pool";
+  if (req.forwarded_once) {
+    if (groups.size() != 1) {
+      authority = req.orig_authority;
+      path = req.orig_path;
     }
   } else {
+    if (faddr_->sni_fwd) {
+      authority = sni_;
+    } else if (!req.authority.empty()) {
+      authority = req.authority;
+    } else {
+      auto h = req.fs.header(http2::HD_HOST);
+      if (h) {
+        authority = h->value;
+      }
+    }
+
+    // CONNECT method does not have path.  But we requires path in
+    // host-path mapping.  As workaround, we assume that path is
+    // "/".
+    if (!req.regular_connect_method()) {
+      path = req.path;
+    }
+
+    // Cache the authority and path used for the first-time backend
+    // selection because per-pattern mruby script can change them.
+    req.orig_authority = authority;
+    req.orig_path = path;
+    req.forwarded_once = true;
+  }
+
+  // Fast path.  If we have one group, it must be catch-all group.
+  if (groups.size() == 1) {
+    group_idx = 0;
+  } else {
+    group_idx = match_downstream_addr_group(routerconf, authority, path, groups,
+                                            catch_all, balloc);
+  }
+
+  if (LOG_ENABLED(INFO)) {
+    CLOG(INFO, this) << "Downstream address group_idx: " << group_idx;
+  }
+
+  if (groups[group_idx]->shared_addr->redirect_if_not_tls && !conn_.tls.ssl) {
+    if (LOG_ENABLED(INFO)) {
+      CLOG(INFO, this) << "Downstream address group " << group_idx
+                       << " requires frontend TLS connection.";
+    }
+    err = SHRPX_ERR_TLS_REQUIRED;
+    return nullptr;
+  }
+
+  auto &group = groups[group_idx];
+  auto addr = get_downstream_addr(err, group.get(), downstream);
+  if (addr == nullptr) {
+    return nullptr;
+  }
+
+  if (addr->proto == Proto::HTTP1) {
+    auto dconn = addr->dconn_pool->pop_downstream_connection();
+    if (dconn) {
+      dconn->set_client_handler(this);
+      return dconn;
+    }
+
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "Downstream connection pool is empty."
                        << " Create new one";
     }
 
-    dconn =
-        make_unique<HttpDownstreamConnection>(group, 0, conn_.loop, worker_);
+    dconn = std::make_unique<HttpDownstreamConnection>(group, addr, conn_.loop,
+                                                       worker_);
+    dconn->set_client_handler(this);
+    return dconn;
   }
 
-  dconn->set_client_handler(this);
+  if (LOG_ENABLED(INFO)) {
+    CLOG(INFO, this) << "Downstream connection pool is empty."
+                     << " Create new one";
+  }
 
+  auto http2session = get_http2_session(group, addr);
+  auto dconn = std::make_unique<Http2DownstreamConnection>(http2session);
+  dconn->set_client_handler(this);
   return dconn;
 }
 
@@ -1154,14 +1007,14 @@ MemchunkPool *ClientHandler::get_mcpool() { return worker_->get_mcpool(); }
 SSL *ClientHandler::get_ssl() const { return conn_.tls.ssl; }
 
 void ClientHandler::direct_http2_upgrade() {
-  upstream_ = make_unique<Http2Upstream>(this);
+  upstream_ = std::make_unique<Http2Upstream>(this);
   alpn_ = StringRef::from_lit(NGHTTP2_CLEARTEXT_PROTO_VERSION_ID);
   on_read_ = &ClientHandler::upstream_read;
   write_ = &ClientHandler::write_clear;
 }
 
 int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
-  auto upstream = make_unique<Http2Upstream>(this);
+  auto upstream = std::make_unique<Http2Upstream>(this);
 
   auto output = upstream->get_response_buf();
 
@@ -1307,7 +1160,7 @@ int ClientHandler::proxy_protocol_read() {
   // NULL character really destroys functions which expects NULL
   // terminated string.  We won't expect it in PROXY protocol line, so
   // find it here.
-  auto chrs = std::array<char, 2>{{'\n', '\0'}};
+  auto chrs = std::array<char, 2>{'\n', '\0'};
 
   constexpr size_t MAX_PROXY_LINELEN = 107;
 
@@ -1478,7 +1331,7 @@ int ClientHandler::proxy_protocol_read() {
   auto &fwdconf = config->http.forwarded;
 
   if ((fwdconf.params & FORWARDED_FOR) &&
-      fwdconf.for_node_type == FORWARDED_NODE_IP) {
+      fwdconf.for_node_type == ForwardedNode::IP) {
     init_forwarded_for(family, ipaddr_);
   }
 
@@ -1488,7 +1341,7 @@ int ClientHandler::proxy_protocol_read() {
 StringRef ClientHandler::get_forwarded_by() const {
   auto &fwdconf = get_config()->http.forwarded;
 
-  if (fwdconf.by_node_type == FORWARDED_NODE_OBFUSCATED) {
+  if (fwdconf.by_node_type == ForwardedNode::OBFUSCATED) {
     return fwdconf.by_obfuscated;
   }
 
